"""
Every frontend provider defines a function named as the module that contains
it, which is to be used as its main entrypoint.  Only one frontend is
typically in use so, unlike with backends, they're not preloaded.
"""
