from setuptools import setup, find_packages

setup(name='blueprint',
      version='3.4.2',
      description='reverse engineer server configuration',
      author='Richard Crowley',
      author_email='richard@devstructure.com',
      url='http://devstructure.com/',
      packages=find_packages(),
      package_data={'blueprint': ['frontend/cfn.json']},
      scripts=['bin/blueprint',
               'bin/blueprint-apply',
               'bin/blueprint-create',
               'bin/blueprint-destroy',
               'bin/blueprint-diff',
               'bin/blueprint-git',
               'bin/blueprint-list',
               'bin/blueprint-prune',
               'bin/blueprint-pull',
               'bin/blueprint-push',
               'bin/blueprint-rules',
               'bin/blueprint-show',
               'bin/blueprint-show-files',
               'bin/blueprint-show-ignore',
               'bin/blueprint-show-packages',
               'bin/blueprint-show-services',
               'bin/blueprint-show-sources',
               'bin/blueprint-split',
               'bin/blueprint-template'],
      license='BSD',
      zip_safe=False)
